import Vec3 from "./utils/Vec3"

// If dragon length is changed, also change it in mainFragment.glsl
const DRAGON_LENGTH = 12


export class DragonSegment {

    pos: Vec3 = new Vec3()

    velocity: Vec3 = new Vec3()    
    
    private temp: Vec3 = new Vec3()
    private temp2: Vec3 = new Vec3()

    constructor(
        private index: number,
        private segmentInFront: DragonSegment | undefined,
        public acceleration: number = 0.01
    ) {}

    update(dragon: Dragon) {

        const timeStep = 1 / 60 // We should run at 60fps, so this should be called 60 times a second

        if (this.segmentInFront === undefined) {
            // Head

            // Get target to follow
            const target: Vec3 = dragon.target

            // Get direction to target
            this.temp.setToVec(target)
            this.temp.sub(this.pos)
            this.temp.normalize()

            // Accelerate towards target
            this.velocity.addScaled(this.temp, this.acceleration * timeStep)

            // Dampen speed
            this.velocity.mul(dragon.airResistance)

            // Update position
            this.pos.addScaled(this.velocity, timeStep)
        }
        else {
            // Keep certain distance from head
            const target: Vec3 = this.segmentInFront.pos

            // Get direction to target
            this.temp.setToVec(target)
            this.temp.sub(this.pos)
            this.temp.normalize()

            // Calculate perfect pos
            this.temp2.setToVec(target)
            this.temp2.addScaled(this.temp, -(dragon.segmentSeparation-(this.index/(DRAGON_LENGTH*1.5))))

            // Fudge position towards ideal distance
            const separationForce = 0.9
            this.pos.mul(1 - separationForce)
            this.pos.addScaled(this.temp2, separationForce)

            // We might want to add some 'inertia' here too?  Or a 'wind' effect?

        }


        // Tell dragon our positions
        const offs = this.index * 3
        dragon.positions[offs+0] = this.pos.x
        dragon.positions[offs+1] = this.pos.y
        dragon.positions[offs+2] = this.pos.z
    }

    reset() {
        this.pos.set(0,0,0)
        this.velocity.set(0,0,0)
    }
}

export default class Dragon {

    segments: DragonSegment[] = []

    positions: number[] = []

    target: Vec3 = new Vec3()

    headAcceleration: number = 20.01
    segmentAcceleration: number = 20.012

    segmentSeparation = 1

    airResistance: number = 0.98 // Just speed dampening

    constructor() {
        // Add segments
        let segmentInFront: DragonSegment | undefined = undefined
        for (let i = 0; i < DRAGON_LENGTH; i++) {
            const acceleration = i == 0 ? this.headAcceleration : this.segmentAcceleration
            const segment: DragonSegment = new DragonSegment(i, segmentInFront, acceleration)
            this.segments.push(segment)
            segmentInFront = segment
        }

        // Pre-allocate positions array
        for (let i =0; i < DRAGON_LENGTH * 3; i++) {
            this.positions.push(0)
        }
    }

    private updateDragon(time: number) {
        for (let segment of this.segments) {
            segment.update(this)
        }
    }

    updateTarget(time: number) {
        // Wave around
        const waveLengthX = 12.23
        const waveLengthY = 4.1
        const waveLengthZ = 17.3
        const dragontime = time -97;
        let targetX = Math.sin(dragontime * Math.PI*2 / waveLengthX + 312.123) * 26.23
        let targetZ = Math.sin(dragontime * Math.PI*2 / waveLengthY + 923.423) * 22.82
        let targetY = Math.sin(dragontime * Math.PI*2 / waveLengthZ + 524.523) * 18.53 + 5

        this.target.set(targetX, targetY, targetZ)
    }

    // Write values to uniform parameters
    updateShaderParameters(time: number, gl: WebGLRenderingContext, program: WebGLProgram) {
        // Update dragon target
        this.updateTarget(time)

        // Update dragon
        this.updateDragon(time)

        // Upload dragon positions
        const dragonPositionUniform = gl.getUniformLocation(program, "dragonPosition")
        gl.uniform3fv(dragonPositionUniform, this.positions)
    }


    // Restart dragon from start
    reset() {
        for(let s of this.segments) {
            s.reset();
        }
    }

}