
import * as utils from "./utils/utils"

// Various parts of the demo are controlled with control signals, that vary over the time of the demo.


export class ControlSignal {

    /**
     * @param name name of the control signal, to be used when passing it to a shader
     * @param timeValuePairs an array with pairs (two-element arrays) of a time in seconds since the start of the demo, and a value for the control signal.  
     *                       The times must be in increasing order.
     * @param fadeShape 0 = jump transformation, 0.5 = smooth transformation, 1 = linear transformation.  Default 0.5
     * @param shaderParameter true if this control signal should be passed to the shader as a float uniform, false if it will not be used in the shader.
     */
    constructor(
        public name: string,
        public timeValuePairs: [time: number, value: number][] = [[0,0]],
        public fadeShape: number = 0.5,
        public shaderParameter: boolean = true
    ) {}

    currentValue: number = 0

    // Calculate the current value for the control signal for the specified time since demo start in seconds
    update(time: number) {
        let index = 0

        let prev = this.timeValuePairs[index]
        let next = this.timeValuePairs[index + 1]

        // Move forward until we get to the correct time bracket
        while (next !== undefined && time >= next[0]) {
            index++
            prev = next
            next = this.timeValuePairs[index + 1]
        }

        if (next === undefined && prev === undefined) {
            // No time value pairs defined, set to zero
            this.currentValue = 0
        }
        else if (next === undefined) {
            // Only one value defined
            this.currentValue = prev?.[1] || 0
        }
        else if (prev === undefined) {
            // Only next value defined??  Shouldn't normally happen
            this.currentValue = next?.[1] || 0
        }
        else {
            // Interpolate
            this.currentValue = utils.smoothMapClamp(time, prev[0], next[0], prev[1], next[1], this.fadeShape)
        }

    }
}


export default class DemoController {

    public controlSignals: ControlSignal[] = []

    constructor() {}

    /**
     * @param name name of the control signal, to be used when passing it to a shader
     * @param timeValuePairs an array with pairs (two-element arrays) of a time in seconds since the start of the demo, and a value for the control signal.  
     *                       The times must be in increasing order.
     * @param fadeShape 0 = jump transformation, 0.5 = smooth transformation, 1 = linear transformation.  Default 0.5
     * @param shaderParameter true if this control signal should be passed to the shader as a float uniform, false if it will not be used in the shader.
     * @returns the created control signal that was added to the controller.  Can be used to access the current value of the control signal if necessary.
     */
    addControlSignal(
        name: string,
        timeValuePairs: [time: number, value: number][] = [[0,0]],
        fadeShape: number = 0.5,
        shaderParameter: boolean = true
    ): ControlSignal {
        const controlSignal = new ControlSignal(name, timeValuePairs, fadeShape, shaderParameter)

        this.controlSignals.push(controlSignal)

        return controlSignal
    }

    // Calculates new values for all control signals for the specified time since demo start in seconds
    update(time: number) {
        for (let signal of this.controlSignals) {
            signal.update(time)
        }
    }

    // Write values to uniform parameters
    updateShaderParameters(gl: WebGLRenderingContext, program: WebGLProgram) {
        for (let signal of this.controlSignals) {
            if (signal.shaderParameter) {
                // Set uniform float parameter of the shader with the name of the signal
                const uniformLocation = gl.getUniformLocation(program, signal.name)
                gl.uniform1f(uniformLocation, signal.currentValue)
            }
        }
    }
}

