
/**
 * Common information for notes.
 * Specific instruments might have unique parameters, in that class create a subclass of the note for that instrument.
 * Others, like plucked strings, need not have any information in the note.
 * 
 * Note instances can be re-used, so should not contain any state information that is modified when playing.
 */
export abstract class Note {


}



// Something that can produce sound samples
export default abstract class Instrument<N extends Note> {

    /**
     * Instrument specific volume multiplier.  1 = default volume.
     */
    public volume: number = 1

    private upcomingNotes: [time: number, note: N][]= []

    /**
     * Schedule a note to be played.
     * 
     * Note: This note should be later than any previously added notes for this instrument
     * 
     * @param noteTime number of seconds since the start of the song
     * @param note the note to be scheduled.  May be a copy of a previous note.
     */
    addNote(noteTime: number, note: N) {
        this.upcomingNotes.push([noteTime, note])
    }

    /**
     * Returns the time when the next note will play, or undefined if there are no scheduled notes
     */
    nextNoteTime(): number | undefined {
        if (this.upcomingNotes.length <= 0) return undefined
        else return this.upcomingNotes[0]?.[0]
    }

    /**
     * Returns the time when the next last added note will play, or undefined if there are no scheduled notes
     */
    lastNoteTime(): number | undefined {
        if (this.upcomingNotes.length <= 0) return undefined
        else return this.upcomingNotes[this.upcomingNotes.length - 1]?.[0]
    }

    /**
     * Gets the next note to be played and removes it from the schedule, or returns undefined if there are none.
     */
    private nextNote(): N | undefined {
        return this.upcomingNotes.shift()?.[1]
    }

    /**
     * Generate sound from the instrument and write it to target buffer.
     * The samples produced should be numbers in the -1 to 1 range by default.
     * 
     * This looks at scheduled notes and triggers them as needed.  This method should not be overwritten in instruments normally.
     * 
     * @param timeSeconds the start time of the sound generation in seconds since music start.
     * @param samplesPerSecond number of samples played each second.
     * @param outputBuffer target bugger to write to
     * @param startSample sample in target buffer to start writing from
     * @param endSample one beyond last sample to write to
     */
    generate(timeSeconds: number, samplesPerSecond: number, outputBuffer: Float32Array, startSample: number, endSample: number): void {

        let currentTime = timeSeconds
        const sampleCount = endSample - startSample
        const endTimeSeconds = timeSeconds + samplesPerSecond * sampleCount
        let currentSample = startSample        

        // Generate sound and trigger notes as needed
        while (currentSample < endSample) {
            // Deterine the time of the next note if we have one
            let nextNoteTime = this.nextNoteTime()
            const nextTime = nextNoteTime !== undefined ? nextNoteTime : endTimeSeconds
            const nextSample = nextNoteTime !== undefined ? currentSample + (nextNoteTime - currentTime) * samplesPerSecond : endSample

            // Generate sound until the next note or the end
            if (nextSample > currentSample) {
                const nextSampleOrEnd = Math.min(nextSample, endSample)
                this.generateSound(currentTime, samplesPerSecond, outputBuffer, currentSample, nextSampleOrEnd)
                currentSample = nextSampleOrEnd
                currentTime = Math.min(nextTime, endTimeSeconds)
            }

            // Check if we should generate a note
            if (nextNoteTime !== undefined && nextSample <= endSample) {
                this.triggerNote(this.nextNote()!)
            }
        }
    }

    /**
     * Handle the triggering of a note.
     * After this, when generateSound is called, the triggered note should be playing.
     * 
     * @param note the note that should start playing
     */
    abstract triggerNote(note: N): void

    /**
     * Generate sound from the instrument and write it to target buffer.
     * The samples produced should be numbers in the -1 to 1 range by default.
     * 
     * This needs to be implemented for each instrument.
     * 
     * @param timeSeconds the start time of the sound generation in seconds since music start.
     * @param samplesPerSecond number of samples played each second.
     * @param outputBuffer target bugger to write to
     * @param startSample sample in target buffer to start writing from
     * @param endSample one beyond last sample to write to
     */
    abstract generateSound(timeSeconds: number, samplesPerSecond: number, outputBuffer: Float32Array, startSample: number, endSample: number): void

    /**
     * Remove all current notes
     */
    reset() {
        this.upcomingNotes = []
    }
}

