import StringInstrument, { StringNote } from "./instruments/StringInstrument";
import WoodwindInstrument, { WoodwindNote } from "./instruments/WoodwindInstrument";
import type Synth from "./Synth";
import * as utils from '../utils/utils'
import GuitarInstrument, { GuitarNote } from "./instruments/GuitarInstrument";

const NOTE_FREQS = {
    "C2": 65.41,
    "C#2": 69.30,
    "D2": 73.42,
    "E2": 82.41,
    "F2": 87.31,
    "G2": 98.00,
    "A2": 110.00,
    "B2": 123.47,
    "C3": 130.81,
    "C#3": 138.59,
    "D3": 146.83,
    "E3": 164.81,
    "F3": 174.61,
    "F#3": 185.00,
    "G3": 196.00,
    "A3": 220.00,
    "B3": 246.94,
    "C4": 261.63,
    "C#4": 277.18,
    "D4": 293.66,
    "E4": 329.63,
    "F4": 349.23,
    "F#4": 369.99,
    "G4": 392.00,
    "A4": 440.00,
    "B4": 493.88,
    "C5": 523.25,
    "C#5": 554.37,
    "D5": 587.33,
    "E5": 659.25,
    "F5": 698.46,
    "F#5": 739.99,
    "G5": 783.99,
    "A5": 880.00,
    "B5": 987.77,
}

type Pitch = keyof typeof NOTE_FREQS
type Chord = "G" | "C" | "Em" | "D" | "A" | "B"

const ARPEGGIOS: {[key in Chord]?: ["" | Pitch, number][]} = {
    //"G": ["B3", "G3", "D4", "G3", "B3", "G3", "B3", "D3"],
    //"C": ["E3", "C3", "G3", "C3", "E3", "C3", "E3", "G2"],
    //"Em": ["G3", "E3", "B3", "E3", "G3", "E3", "G3", "B2"],
    //"D": ["F#3", "D3", "A3", "D3", "F#3", "D3", "F#3", "A2"],
    "D": [
        ["D4", 1],
        ["A3", 1],
        ["F#3", 2],
    ],
    "A": [
        ["C#4", 1],
        ["A3", 1],
        ["E3", 2],
    ],
    "B": [
        ["B3", 1],
        ["F#3", 1],
        ["D3", 2],
    ],
    "G": [
        ["B3", 1],
        ["G3", 1],
        ["D3", 2],
    ],
}

const PROGRESSION: Chord[] = ["D", "A", "B", "G"]
/*[
    ["G4", 1],
    ["F4", 0.5],
    ["D4", 0.5],
    ["G4", 0.5],
    ["C5", 0.5],
    ["G4", 1],
],
[
    ["E5", 0.5],
    ["D5", 0.5],
    ["C5", 1],
    ["G4", 0.5],
    ["F4", 0.5],
    ["E4", 1],
],
[
    ["E5", 1],
    ["D5", 0.5],
    ["B4", 0.5],
    ["E5", 0.5],
    ["A5", 0.5],
    ["E5", 1],
],
[
    ["D5", 0.5],
    ["", 0.5],
    ["C5", 0.5],
    ["", 0.5],
    ["A4", 2],
]*/

type Melody =  ["" | Pitch, number][];

const MELODY1: Melody[] = [
    [
        // D
        ["D5", 0.75],
        ["F#5", 0.5],
        ["F#5", 0.5],
        ["", 2.25],
    ],
    [
        // A
        ["A4", 0.75],
        ["D5", 0.5],
        ["D5", 0.5],
        ["", 2.25],
    ],
    [
        // B
        ["B4", 0.75],
        ["D5", 0.5],
        ["D5", 0.5],
        ["", 2.25],
    ],
    [
        // G
        ["B4", 0.75],
        ["G4", 0.5],
        ["G4", 0.5],
        ["", 2.25],
    ],
];
const MELODY2: Melody[] = [
    [
        // D
        ["", 4],
    ],
    [
        // A
        ["", 4],
    ],
    [
        // B
        ["", 4],
    ],
    [
        // G
        ["", 2],
        ["G4", 1],
        ["B4", 1],
    ],
];
const MELODY3: Melody[] = [
    [
        // D
        ["A4", 3],
        ["D5", 1],
    ],
    [
        // A
        ["C#5", 0.5],
        ["B4", 0.5],
        ["A4", 0.5],
        ["G4", 0.5],
        ["F#4", 2],
    ],
    [
        // B
        ["B4", 3],
        ["E5", 1],
    ],
    [
        // G
        ["G5", 0.5],
        ["F#5", 0.5],
        ["E5", 0.5],
        ["D5", 0.5],
        ["B4", 2],
    ],
    [
        // D
        ["F#5", 1],
        ["D5", 1],
        ["A4", 2],
    ],
    [
        // A
        ["E5", 1],
        ["C#5", 1],
        ["", 1],
        ["E5", 1],
    ],
    [
        // B
        ["D5", 1],
        ["C#5", 1],
        ["B4",  2],
    ],
    [
        // G
        ["G4", 2],
        ["", 2],
    ]
];

const MELODY = MELODY1.concat(MELODY1,  MELODY2, MELODY3, MELODY3);

export default class Music {

    guitar: GuitarInstrument

    patternLength = 3
    nextPatternStart = 0
    fadeInLength = 4


    /**
     * @param synth the class that provides sound functionality.  Accessible in this class with this.synth.
     */
    constructor(private synth: Synth) {

        const guitarMasterVolume = 1
        const strings = 3 // Let's keep it low to keep up performance
        this.guitar = new GuitarInstrument(strings, 50, this.synth.samplesPerSecond, guitarMasterVolume, synth)
        this.synth.addInstrument(this.guitar)
    }


    woodwind = this.synth.addInstrument(new WoodwindInstrument(this.synth.samplesPerSecond))


    setup() {
    }


    // Pattern based music generation
    createPattern(time: number, duration: number) {
        const endTime = time + duration
        const patternId = Math.floor(time / this.patternLength) % MELODY.length

        // Generate melody
        if (patternId >= 0) {
            const melody = MELODY[patternId]!
            let melodyTime = time
            for (const [pitch, length] of melody) {
                // Add note
                if (pitch !== "") {
                    this.woodwind.addNote(melodyTime, new WoodwindNote(0.6, NOTE_FREQS[pitch], length / 4 * duration))
                }
                
                melodyTime += length / 4 * duration
            }
        }
        

        // Generate arpeggio
        const chord = PROGRESSION[patternId % 4]!
        const arpeggio = ARPEGGIOS[chord]!.concat(ARPEGGIOS[chord]!)
        let arpeggioTime = time
        for (const [pitch, length] of arpeggio) {
            if (pitch !== "") {
                this.guitar.addNote(arpeggioTime, new GuitarNote(NOTE_FREQS[pitch], 0.5, 0.2))
            }
            arpeggioTime += length / 4 * duration
        }
    }


    /**
     * Called whenever the audio buffer needs to be filled in, which is about ten times a second.
     * However, do not rely on exact call frequency.
     * 
     * @param time seconds since the start of the music
     * @param duration number of seconds until the next update will be called.
     */
    update(time: number, duration: number) {
        // Fade in
        const fadeShape = 0.8 // 0.5 smooth, 0 snap, 1 linear
        const volumeFade = utils.smoothMap(time, 0.2 * this.fadeInLength, this.fadeInLength, 0, 1, fadeShape)
        this.synth.masterVolume = volumeFade

        // How many seconds in advance to create the next pattern
        const margin = this.patternLength/2

        // Create next pattern
        if (time + margin >= this.nextPatternStart ) {
            this.createPattern(this.nextPatternStart, this.patternLength)

            this.nextPatternStart += this.patternLength
        }

    }

    /**
     * Restart music from start
     */
    reset() {
        this.nextPatternStart = 0
    }

}
