import type { Note } from "./Instrument"
import type Instrument from "./Instrument"
import type Music from "./Music"


export default class Synth {

    public time = 0

    public readonly samplesPerSecond: number
    public readonly secondsPerSample: number

    public instruments: Instrument<any>[] = []

    public masterVolume: number = 1

    private scriptProcessorNode: ScriptProcessorNode | undefined = undefined

    private music: Music | undefined = undefined

    // Adjust all sound volume with this, it's easy for a demo to be really loud on the main stage
    private masterVolumeTune: number = 0.8

    audioContext: AudioContext
    tempBuffer: Float32Array

    constructor(public readonly bufferSize = 4096) {
        // Create context
        this.audioContext = new window.AudioContext();

        this.tempBuffer = new Float32Array(new ArrayBuffer(bufferSize * 4))


        // Init parameters
        this.samplesPerSecond = this.audioContext.sampleRate
        this.secondsPerSample = 1 / this.audioContext.sampleRate
    }

    /**
     * @param instrument instrument to add to the synth.  It means that sound will be generated using it.
     * @returns the added instrument, for possible chaining.
     */
    addInstrument<N extends Note, T extends Instrument<N>>(instrument: T): T {
        this.instruments.push(instrument)
        return instrument
    }


    /** 
     * Start playing sound
     * @param music the song or procedural music to play.  It's update callback will be called whenever the buffer is being refilled.
     */
    start(music: Music) {
        this.music = music

        this.time = 0

        const stereo = false

        // Setup music
        music.setup()
        
        // Setup soundcallback
        this.scriptProcessorNode = this.audioContext.createScriptProcessor(this.bufferSize, 1, stereo ? 2 : 1);
        this.scriptProcessorNode.onaudioprocess = (e: AudioProcessingEvent) => {

            // @ts-ignore    Only runs in development and will be stripped in production builds.
            if (process.env.NODE_ENV !== 'production') {
                // Time audio generation duration
                performance.mark("sound-start")
            }

            const startSample = 0
            const endSample = this.bufferSize
            const sampleCount = endSample - startSample;

            let duration = sampleCount * this.secondsPerSample

            // Music generation callback
            music.update(this.time, duration)


            let outputBuffer: Float32Array = e.outputBuffer.getChannelData(0)
            //bufferLeft = e.outputBuffer.getChannelData(0)
            //bufferRight = e.outputBuffer.getChannelData(1)


            // Clear out buffer
            outputBuffer.fill(0, startSample, endSample)

            // Generate instrument sounds
            for (let instrument of this.instruments) {
                // Clear temp buffer
                this.tempBuffer.fill(0, startSample, endSample)

                // Generate instrument sound to temp buffer
                instrument.generate(this.time, this.samplesPerSecond, this.tempBuffer, startSample, endSample)

                // Add temp buffer to output buffer, and scale volume
                const volumeScale = instrument.volume * this.masterVolume * this.masterVolumeTune;
                for (let i = startSample; i < endSample; i++) {
                    outputBuffer[i] += this.tempBuffer[i]! * volumeScale
                }
            }


            // Update time
            this.time += duration
            
            /*
            // Clamp output buffer to range
            for (let i = startSample; i < endSample; i++) {
                const v: number = outputBuffer[i]!
                if (v < -1) outputBuffer[i] = -1
                else if (v > 1) outputBuffer[i] = 1
            }
            */

            // @ts-ignore    Only runs in development and will be stripped in production builds.
            if (process.env.NODE_ENV !== 'production') {
                // Time audio generation duration
                performance.mark("sound-end")
            }
        }

        this.scriptProcessorNode.connect(this.audioContext.destination)
    }

    stop() {
        this.audioContext.close()
    }

    restart() {
        this.time = 0

        // Reset instruments to clear out queued notes
        for (let instr of this.instruments) {
            instr.reset()
        }

        this.music?.reset()
    }
}





