import Instrument from "../Instrument"
import type Synth from "../Synth"
import StringInstrument from "./StringInstrument"

/**
 * Parameters describing how a string is plucked
 */
export class GuitarNote {

    /**
     * @param frequency the frequency to play the string with
     * @param pluckStrength how strong noise to add to the string, 1 = normal full strength.
     * @param dampenPreviousSound number to multiply previous string state with
     */
    constructor(
        public frequency: number = 220,
        public pluckStrength: number = 1,
        public dampenPreviousSound: number = 0.2,
    ) {
    }
}


/**
 * A Karplus-Strong waveguide instrument, for guitar-like sounds.
 */
export default class GuitarInstrument extends Instrument<GuitarNote> {

    // Strings
    strings: StringInstrument[] = []

    tempBuffer: Float32Array

    /**
     * @param {number} numberOfStrings number of strings that the instrument has.  New notes are played on the string with currently lowest volume.
     * @param {number} minFrequency minimum frequency that can be played with this instrument (how 'long' it can be at most, how much space to allocate to the wavetable)
     * @param {number} samplesPerSecond number of samples per second used.
     * @param {number} volume volume adjustment for this string.  1 = normal, 0 = fully muted
     * @param {number} stretchFactor At 0, sounds like before, at 0.5, stretches sound quite a bit and makes it more electrical.  At 1, things break.
     */
    constructor(
        private numberOfStrings: number = 6,
        private minFrequency: number = 40,
        private samplesPerSecond: number,
        volume: number = 1,
        synth: Synth,
        public stretchFactor: number = 0.25
    ) {
        super()

        this.volume = volume

        // Allocate temp buffer for adding string sound
        this.tempBuffer = new Float32Array(new ArrayBuffer(synth.bufferSize * 4))


        // Create strings
        for (let i = 0; i < this.numberOfStrings; i++) {
            this.strings.push(new StringInstrument(minFrequency, samplesPerSecond, 1, stretchFactor))
        }
    }


    /**
     * Pluck the string
     * @param note information about the pluck
     */
    override triggerNote(note: GuitarNote): void {

        // Find the most quiet string
        let quietest = this.strings[0]
        let quietestAmplitude = this.strings[0]?.measuredAmplitude || 100
        for (let s of this.strings) {
            let amp = s.measuredAmplitude
            if (amp < quietestAmplitude) {
                quietest = s
                quietestAmplitude = amp
            }
        }

        // Trigger the quietest string
        quietest?.triggerNote(note)
    }

    /**
     * Generate sound from the waveguide and write it to target buffer
     * 
     * @param timeSeconds the start time of the sound generation in seconds since music start.
     * @param samplesPerSecond number of samples played each second.
     * @param outputBuffer target bugger to write to
     * @param startSample sample in target buffer to start writing from
     * @param endSample one beyond last sample to write to
     */
    override generateSound(timeSeconds: number, samplesPerSecond: number, outputBuffer: Float32Array, startSample: number, endSample: number): void {

        // Add sounds of strings
        for (let string of this.strings) {
            // Clear temp buffer
            this.tempBuffer.fill(0, startSample, endSample)

            // Generate string sound to temp buffer
            string.generate(timeSeconds, samplesPerSecond, this.tempBuffer, startSample, endSample)

            // Add temp buffer to output buffer, and scale volume
            const volumeScale = string.volume * this.volume;
            for (let i = startSample; i < endSample; i++) {
                outputBuffer[i] += this.tempBuffer[i]! * volumeScale
            }
        }
        
    }
}
