import Instrument from "../Instrument"

/**
 * Parameters describing how a string is plucked
 */
export class StringNote {

    /**
     * @param frequency the frequency to play the string with
     * @param pluckStrength how strong noise to add to the string, 1 = normal full strength.
     * @param dampenPreviousSound number to multiply previous string state with
     */
    constructor(
        public frequency: number = 220,
        public pluckStrength: number = 1,
        public dampenPreviousSound: number = 0.2,
    ) {
    }
}


/**
 * A Karplus-Strong waveguide instrument, for guitar-like sounds.
 */
export default class StringInstrument extends Instrument<StringNote> {
    // Wavetable
    wavetable: Float32Array
    wavetableLength: number

    // Sample pos
    updatePos: number = 0

    // How much of the string is currently in use
    currentWavetableLength: number = 0

    previousSample: number = 0

    // Measure approximate maximum amplitude
    public measuredAmplitude: number = 0

    /**
     * @param {number} minFrequency minimum frequency that can be played with this string (how 'long' it can be at most, how much space to allocate to the wavetable)
     * @param {number} samplesPerSecond number of samples per second used.
     * @param {number} volume volume adjustment for this string.  1 = normal, 0 = fully muted
     */
    constructor(
        private minFrequency: number,
        private samplesPerSecond: number,
        volume: number = 1,
        private stretchFactor: number = 0.4
    ) {
        super()

        this.volume = volume

        // Allocate suitable length so that resonance forms
        this.wavetableLength = Math.round(samplesPerSecond / minFrequency);
        this.wavetable = new Float32Array(new ArrayBuffer(this.wavetableLength * 4))

        // Starting value
        this.currentWavetableLength = this.wavetableLength

        // Reset wavetable
        this.wavetable.fill(0);
    }


    /**
     * Pluck the string
     * @param note information about the pluck
     */
    override triggerNote(note: StringNote): void {
        const frequency = note.frequency
        const volume = note.pluckStrength
        const dampen = note.dampenPreviousSound

        this.measuredAmplitude = 1

        // Adjust length of wavetable that we use based on frequency of the note we play
        this.currentWavetableLength = Math.min(Math.round(this.samplesPerSecond / frequency), this.wavetableLength);

        // Fill free part of string with noise
        for (let i = 0; i < this.currentWavetableLength; i++) {
            // Simple random -1 to 1 noise:
            this.wavetable[i] = dampen * this.wavetable[i]! + volume * (2 * Math.random() - 1)
        }

        // Fill rest of string with the sound of silence
        for (let i = this.currentWavetableLength; i < this.wavetableLength; i++) {
            this.wavetable[i] = 0
        }
    }

    /**
     * Generate sound from the waveguide and write it to target buffer
     * 
     * @param timeSeconds the start time of the sound generation in seconds since music start.
     * @param samplesPerSecond number of samples played each second.
     * @param outputBuffer target bugger to write to
     * @param startSample sample in target buffer to start writing from
     * @param endSample one beyond last sample to write to
     */
    override generateSound(timeSeconds: number, samplesPerSecond: number, outputBuffer: Float32Array, startSample: number, endSample: number): void {

        // Fill output buffer
        for (let i = startSample; i < endSample; i++) {
            
            let sample = this.wavetable[this.updatePos]!

            // Use stretching factor to pull out sound for longer time by sometimes skipping the averaging, which is useful especially for high notes
            if (Math.random() >= this.stretchFactor) {
                // Average current and previous position in the wavetable, save into current position
                sample =  0.5 * (sample + this.previousSample)
                this.wavetable[this.updatePos] = sample
            }

            this.previousSample = sample

            // Output sample
            outputBuffer[i] = sample

            // Move to next sample, with wrap-around at the point where the string is 'pinched'
            this.updatePos++;
            if (this.updatePos >= this.currentWavetableLength) this.updatePos = 0;

            // Measure amplitude
            const amplitudeMeasureDampening = 0.999
            this.measuredAmplitude = Math.max(this.measuredAmplitude * amplitudeMeasureDampening, Math.abs(sample))
        }


    }
}
