import Instrument from "../Instrument";

// Parameters
export class WoodwindNote {
    constructor(
        public volume: number = 1,
        public frequency: number = 440,
        public length: number = 1,
    ) {
    }
}

class PlayingNote {
    dead = false;

    constructor(
        public note: WoodwindNote,
        public pos: number,
        private samplesPerSecond: number,
    ) {
    }

    get time() {
        return this.pos / this.samplesPerSecond
    }

}

const attackTime = 0.1; // Attack time in seconds
const decayTime = 0.1; // Decay time in seconds
const releaseTime = 0.2; // Release time in seconds
const sustainLevel = 0.7; // sustain level

/**
 * A simple woodwind instrument.
 */
export default class WoodwindInstrument extends Instrument<WoodwindNote> {
    playingNotes: PlayingNote[] = []

    constructor(private samplesPerSecond: number) {
        super()
    }

    override triggerNote(note: WoodwindNote): void {
        this.playingNotes.push(new PlayingNote(note, 0, this.samplesPerSecond))
    }

    override generateSound(timeSeconds: number, samplesPerSecond: number, outputBuffer: Float32Array, startSample: number, endSample: number): void {

        this.playingNotes = this.playingNotes.filter(n => !n.dead);
        
        // Clear buffer
        for (let i = startSample; i < endSample; i++) outputBuffer[i] = 0;

        // Iterate all notes playing
        for (const playingNote of this.playingNotes) {
            const fundamentalFrequency = playingNote.note.frequency;
            const harmonics = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]; // fundamental and harmonics
            const offsets = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1];
            const amplitudes = [1, 0.5, 0.25, 0.1, 0.05, 0.025, 0.0125, 0.00625, 0.003125, 0.0015625]; // example amplitudes
            const amplitudeSum = amplitudes.reduce((a, b) => a + b, 0);

            for (let i = startSample; i < endSample; i++) {
                let sample = 0;
                playingNote.pos++;

                // Add sine waves for each harmonic
                for (let h = 0; h < harmonics.length; h++) {
                    const frequency = fundamentalFrequency * harmonics[h]!;
                    const t = playingNote.time + offsets[h]!;
                    sample += Math.sin(2 * Math.PI * frequency * t) * amplitudes[h]!;
                }

                // Normalize sample - depending on the sum of amplitudes; this may not be needed if the amplitudes are already normalized
                sample /= amplitudeSum;

                // Apply a simple envelope to simulate attack and release

                let envelope;

                // Apply envelope
                if (playingNote.time < attackTime) {
                    envelope = playingNote.time / attackTime * playingNote.note.volume;
                } else if (playingNote.time < attackTime+decayTime) {
                    const d = (playingNote.time - attackTime) / decayTime;
                    envelope = d * (sustainLevel * playingNote.note.volume);
                    envelope += (1-d) * playingNote.note.volume;
                } else if (playingNote.time < playingNote.note.length - releaseTime) {
                    envelope = sustainLevel * playingNote.note.volume;
                } else if (playingNote.time < playingNote.note.length) {
                    envelope = (1 - (playingNote.time - playingNote.note.length + releaseTime) / releaseTime) * sustainLevel * playingNote.note.volume;
                } else {
                    envelope = 0;
                    playingNote.dead = true;
                }

                // Assign the sample to the buffer, applying the envelope
                outputBuffer[i] += sample * envelope;
            }
        }
    }
}
