
/**
 * Returns pseudorandom number that depends on the current time and optionally the provided seed value.
 * Not high quality random, but compact, depends on time and the sample, and the default javascript random doesn't support seeds.
 * Returns a value in the 0..1 range.
 * @param {number} seed 
 * /
function random(seed = 0) {
    var x = Math.sin(t + sampleTime * 7 + seed * 97) * 10000;
    return x - Math.floor(x);
}
*/


/**
 * Uses smootherstep interpolation ( https://en.wikipedia.org/wiki/Smoothstep#Variations )
 * @param {number} v value between 0 and 1, that should be smoothed to value between 0 and 1.
 * @param {number} s shape parameter, 0 = jump transformation, 0.5 = smooth transformation, 1 = linear transformation.
 * @noinline
 */
export function smooth(v: number, s = 0.5) {
    let x = +v || 0;
    let p = x * x * x * (x * (x * 6 - 15) + 10); // Smoothed transition
    return x <= 0 ? 0 : x >= 1 ? 1 : +s <= 0.5 ? mix(x < 0.5 ? 0 : 1, p, +s * 2) : mix(p, x, (+s - 0.5) * 2); // For 0..0.5 range mix between jump and smooth, for 0.5..1 range mix between smooth and linear.
  }
  
  /**
   * Interpolates between a and b, uses a when t = 0 and b when t = 1.
   */
  export function mix(a: number, b: number, t: number) {
    return a + (b - a) * t;
  }
  
  /**
   * Maps between targetStart and targetEnd, when t is between sourceStart and sourceEnd.
   * Smooths value
   * @param {number} shape shape parameter, 0 = jump transformation, 0.5 = smooth transformation, 1 = linear transformation.
   * @noinline
   */
  export function smoothMap(t: number, sourceStart = 0, sourceEnd = 1, targetStart = 0, targetEnd = 1, shape = 0.5) {
    let r = (sourceEnd == sourceStart ? 0.5 : (t - sourceStart) / (sourceEnd - sourceStart)) || 0;
    r = smooth(r, shape);
    return (targetStart + (targetEnd - targetStart) * r) || 0; // Ensure we return a number
  }
  
  /**
   * Maps between targetStart and targetEnd, when t is between sourceStart and sourceEnd.
   * Smooths value, and clamps value
   * @param {number} shape shape parameter, 0 = jump transformation, 0.5 = smooth transformation, 1 = linear transformation.
   * @noinline
   */
  export function smoothMapClamp(t: number, sourceStart = 0, sourceEnd = 1, targetStart = 0, targetEnd = 1, shape = 0.5) {
    let r = (sourceEnd == sourceStart ? 0.5 : (t - sourceStart) / (sourceEnd - sourceStart)) || 0;
    if (r < 0) r = 0;
    if (r > 1) r = 1;
    r = smooth(r, shape);
    return (targetStart + (targetEnd - targetStart) * r) || 0; // Ensure we return a number
  }
  

/**
 * Maps between targetStart and targetEnd, when t is between sourceStart and sourceEnd.
 * Clamps values to the range.
 * @noinline
 */
export function mapClamp(t: number, sourceStart: number = 0, sourceEnd: number = 1, targetStart: number = 0, targetEnd: number = 1) {
    let r = (sourceEnd == sourceStart ? 0.5 : (t - sourceStart) / (sourceEnd - sourceStart));
    if (r < 0) r = 0;
    else if (r > 1) r = 1;
    return (targetStart + (targetEnd - targetStart) * r)
}
  

export function clamp(value: number, minValue: number, maxValue: number): number {
    return value < minValue ? minValue : (value > maxValue ? maxValue : value)
}

// Make specified element fullscreen.  Needs to be triggered by button.
export function requestFullScreen(element: HTMLElement) {
    if (element.requestFullscreen) {
        element.requestFullscreen();
    } else if (element.webkitRequestFullscreen) { /* Safari */
        element.webkitRequestFullscreen();
    } else if (element.msRequestFullscreen) { /* IE11 */
        element.msRequestFullscreen();
    }
}
