#version 300 es
precision highp float;

//===============================================================
//    Shared Structures
//===============================================================

//struct SnowstormParameters {
//};


//===============================================================
//    Shader Parameters
//===============================================================

// Inputs
uniform float time; // Time in seconds since start
uniform vec2 screenSize; // Screen size in pixels
uniform uint frameNum; // Frame number since start

// Control signals
uniform float fadeFromBlack; // Fade from black at start.  0 = black, 1 = normal

uniform float dragonOnScreen;

uniform float growingDropletRadiusScale;
uniform float growingSceneOn;
uniform float growingFlakeNumber;
uniform float growingHexSizeScale;
uniform float growingHalfFlakeScale;
uniform float growingFlakeScale;
uniform float growingDropletYOffset;
uniform float growingFlakerot;
uniform float growingBaserotmod;
uniform float dragonXrotMod;
uniform float snowstormStart;
uniform float stormStartScale;

// Dragon 
// If dragon length is changed, also change it in Dragon.ts
#define DRAGON_LENGTH 12
uniform vec3 dragonPosition[DRAGON_LENGTH]; // Position of each segment of the dragon

// Outputs (color at pixel)
out vec4 FragColor;

//===============================================================
//     Constants
//===============================================================

#define MAX_RAY_STEPS 50
#define MAX_SHADOW_RAY_STEPS 20
#define MAX_DISTANCE 1000.
#define SURFACE_DISTANCE .001
#define STEP_DISTANCE_MULTIPLIER 1.0

#define PI 3.14159265359

const float TAU = 2. * PI;
const float BIG_ENOUGH = MAX_DISTANCE+10.0;

//===============================================================
//     Commonly Used Data Structures
//===============================================================


// Data returned for a specific point in a scene.
// Includes the distance to the closest surface (negative if inside the surface),
// the id of the material (of the closest surface, -1 indicates sky/no object), and a material parameter that can 
// depend on the object in some way (default to 0).
struct SceneSample {
    float surfaceDistance;
    int materialId;
    float materialParameter;
};



//===============================================================
//     Rendering parameters
//===============================================================


// Sunlight
vec3 sunDir = normalize(vec3(-1., 1., 3.));
vec3 sunColor = vec3(0.8863, 0.6, 0.4471);
float sunIntensity = 2.0;
vec3 sunShadowPower = vec3(1.0, 1.5, 1.2); // Tint the shadow a bit by raising it to a power.

// Skylight
vec3 skyColor = vec3(0.1412, 0.1804, 0.2941);
float skyIntensity = 0.7;

// Indirect light (sunlight bouncing from scene, or light of skydome opposite sun)
vec3 indirectColor = vec3(0.3961, 0.1176, 0.3804);
float indirectIntensity = .5;

// Atmosphere
const float fogIntensity = 0.04;

// When the fog covers more than this, skip rendering any underlying scene details.
const float fogExtinctionThreshold = 0.98;


//===============================================================
//     Math Utilities
//===============================================================

// Maps value from input range to target range and clamps to target range.
// Does not handle cases where input range length is zero (results in division by zero instead of middle of output range).
float mapClamp(float value, float inStart, float inEnd, float outStart, float outEnd) {
  return outStart + clamp(((value - inStart) / (inEnd - inStart)), 0., 1.) * (outEnd - outStart);
}


//===============================================================
//     Projection
//===============================================================

vec2 screenToWorld(vec2 screenPos) {
    vec2 result = 2.0 * (screenPos / screenSize.xy - 0.5);
    result.x *= screenSize.x / screenSize.y; // Correct width by aspect ratio (height remains -1 to 1)
    return result;
}

vec3 getCameraRayDir(vec2 uv, vec3 camPos, vec3 camTarget) {

    // Camera field of view / zoom.
    float zoom = 2.0;  // sin(time) * 0.5 + 2.;

    // Calculate camera's "orthonormal basis", i.e. its transform matrix components
    vec3 camForward = normalize(camTarget - camPos);
    vec3 camRight = normalize(cross(vec3(0.0, 1.0, 0.0), camForward));
    vec3 camUp = normalize(cross(camForward, camRight));

    // Camera ray direction
    return normalize(uv.x * camRight + uv.y * camUp + camForward * zoom);
}


//===============================================================
//     Utils
//===============================================================

float dot2(in vec2 v ) { return dot(v,v); }

//modified version of iq code to extrude 2d object to 3d
// h seems to be the 'radius' of the height, extruded that much up and down
float opExtrusion( in vec3 p, in float sdf, in float h )
{
    vec2 w = vec2( sdf, abs(p.z) - h );
    return min(max(w.x,w.y),0.0) + length(max(w,0.0));
}

//code for 2d hexagon copied from iq
float sdHexagon( in vec2 p, in float r )
{
    const vec3 k = vec3(-0.866025404,0.5,0.577350269);
    p = abs(p);
    p -= 2.0*min(dot(k.xy,p),0.0)*k.xy;
    p -= vec2(clamp(p.x, -k.z*r, k.z*r), r);
    return length(p)*sign(p.y);
}

//code for 2d trapezoid from iq
float sdTrapezoid( in vec2 p, in float r1, float r2, float he )
{
    vec2 k1 = vec2(r2,he);
    vec2 k2 = vec2(r2-r1,2.0*he);
    p.x = abs(p.x);
    vec2 ca = vec2(p.x-min(p.x,(p.y<0.0)?r1:r2), abs(p.y)-he);
    vec2 cb = p - k1 + k2*clamp( dot(k1-p,k2)/dot2(k2), 0.0, 1.0 );
    float s = (cb.x<0.0 && ca.y<0.0) ? -1.0 : 1.0;
    return s*sqrt( min(dot2(ca),dot2(cb)) );
}

//rotates any 2d point that it is multiplied with for given angle
mat2 Rot(float a) {
   float s = sin(a);
   float c = cos(a);
    return mat2(c, -s, s, c);
}

//moves 2d point the given offset
vec2 move2d(vec2 p, vec2 offset){
    return p - offset;
}

vec3 move(vec3 p, vec3 position) {
    return p - position;
}

//mirrs along y
vec2 y_mirror2d(vec2 p){    
    return vec2(p.x, -p.y);
}

// Rotate coordinate space by specified angle in degrees.
// TODO: Refactor
vec2 rotate2d(vec2 p, float angle) {
    float a = -radians(angle);
    
    float x = p.x * cos(a) - p.y * sin(a);
    float y = p.x * sin(a) + p.y * cos(a);
    return vec2(x, y);
}




/** Rotate a 2D point by the specified number of turns. */
vec2 rot2(vec2 p, float turns) {
    float s = sin(TAU * turns);
    float c = cos(TAU * turns);
    return vec2(
        p.x * c - p.y * s,
        p.x * s + p.y * c
    );
}


/** Rotate a point around y axis by the specified number of turns. */
vec3 xRot(vec3 p, float turns) {
    vec2 yz = rot2(p.yz, turns);
    return vec3(p.x, yz.x, yz.y);
}

/** Rotate a point around y axis by the specified number of turns. */
vec3 yRot(vec3 p, float turns) {
    vec2 xz = rot2(p.xz, turns);
    return vec3(xz.x, p.y, xz.y);
}

/** Rotate a point around z axis by the specified number of turns. */
vec3 zRot(vec3 p, float turns) {
    vec2 xy = rot2(p.xy, turns);
    return vec3(xy.x, xy.y, p.z);
}

// Rotate space (represented by sampling point p), around the center point, so that the X axis points in the specified direction, and the y axis points approximately upwards.
vec3 place(vec3 p, vec3 center, vec3 direction) {
    vec3 a = normalize(direction);
    vec3 b = normalize(cross(a, vec3(0, 1, 0)));
    vec3 c = normalize(cross(a, b));

    mat3 rot = mat3(-b, -c, -a);

    return rot * (p - center);
}

float sdOctahedron( vec3 p, float s)
{
  p = abs(p);
  return (p.x+p.y+p.z-s)*0.57735027;
}

vec3 up(vec3 p, float d) {
    return p - vec3(0., d, 0.);
}

vec3 right(vec3 p, float d) {
    return p - vec3(0., 0., d);
}

vec3 back(vec3 p, float d) {
    return p - vec3(-d, 0., 0.);
}

// Returns radius in u and angle in v (in radians).
vec2 cartesianToPolar(vec2 p) {
    float radius=length(p);
    float angle=atan(p.y, p.x);
    return vec2(radius, angle);
}

float smin( float a, float b, float k )
{
    k *= 1.0;
    float res = exp2( -a/k ) + exp2( -b/k );
    return -k*log2( res );
}

// Combine two scene samples, return the one that is closest
SceneSample closest(SceneSample a, SceneSample b) {
    if (a.surfaceDistance <= b.surfaceDistance) {
        return a;
    }
    else {
        return b;
    }
}

SceneSample smootClosest(SceneSample a, SceneSample b, float smootParam) {
    float dist = smin(a.surfaceDistance,b.surfaceDistance, smootParam);
    float mixValue = a.surfaceDistance/(a.surfaceDistance+ b.surfaceDistance);
    float materialParameter = mapClamp(mixValue, 0.0, 1.0, a.materialParameter, b.materialParameter);
    return SceneSample(dist, a.materialId, materialParameter);
}


// A classical pseurdorandom function.  Returns a value between 0 and 1 (exclusive).
float random(vec2 p) {
    return fract(sin(dot(p.xy, vec2(12.9898, 78.233))) * 43758.5453);
}

// Feeds an integer vector to random
float random(ivec2 p) {
    // Just naive casting.  Is there a more compact way?
    return random(vec2(float(p.x), float(p.y)));
}

// Returns random value for integer vector and a 'seed' number
float random(ivec2 p, float pseudoSeed) {
    // Just naive casting.  Is there a more compact way?
    return random(vec2(float(p.x) + pseudoSeed * 1953.9023, float(p.y) + pseudoSeed * 7312.7312));
}


const uint hashK = 1103515245U; // As used in GLIB C

// Hash function 3 ints -> 3 floats.  Originally by iq: https://www.shadertoy.com/view/XlXcW4
vec3 hash3( uvec3 x ) {
    x = ((x>>8U)^x.yzx) * hashK;
    x = ((x>>8U)^x.yzx) * hashK;
    x = ((x>>8U)^x.yzx) * hashK;
    
    return vec3(x)*(1.0/float(0xffffffffU));
}

// Hash function 2 ints -> float. Originally by iq: https://www.shadertoy.com/view/4tXyWN
float hash( uvec2 x ) {
    uvec2 q = 1103515245U * ( (x>>1U) ^ (x.yx   ) );
    uint  n = 1103515245U * ( (q.x  ) ^ (q.y>>3U) );
    return float(n) * (1.0/float(0xffffffffU));
}

// pcg hash function.  A compromise on a fast & good hash function.
// See the paper: https://www.jcgt.org/published/0009/03/02/paper.pdf
// And the implementations. https://www.shadertoy.com/view/XlGcRh
//
// See https://www.shadertoy.com/view/ctj3Wc for an adaptation into a rng function with internal state
// 
// https://www.pcg-random.org/
uint pcg(uint v) {
	uint state = v * 747796405u + 2891336453u;
	uint word = ((state >> ((state >> 28u) + 4u)) ^ state) * 277803737u;
	return (word >> 22u) ^ word;
}

uvec2 pcg2d(uvec2 v) {
    v = v * 1664525u + 1013904223u;

    v.x += v.y * 1664525u;
    v.y += v.x * 1664525u;

    v = v ^ (v>>16u);

    v.x += v.y * 1664525u;
    v.y += v.x * 1664525u;

    v = v ^ (v>>16u);

    return v;
}

// Utility to create a simple waveform
float wobble(float wavelen, float average, float variation, float phase) {
    return sin((time * TAU) / wavelen + phase) * variation + average;
}

float wobble(float time, float wavelen, float average, float variation, float phase) {
    return sin((time * TAU) / wavelen + phase) * variation + average;
}


//===============================================================
//     Voronoi
//===============================================================

// TODO: Cleanup

vec2 random2( vec2 p ) {
    return fract(sin(vec2(dot(p,vec2(127.1,311.7)),dot(p,vec2(269.5,183.3))))*43758.5453);
}

// From https://iquilezles.org/articles/voronoilines/
float voronoiDistance( in vec2 x ) {
    ivec2 p = ivec2(floor( x ));
    vec2  f = fract( x );

    ivec2 mb;
    vec2 mr;

    float res = 8.0;
    for( int j=-1; j<=1; j++ )
    for( int i=-1; i<=1; i++ )
    {
        ivec2 b = ivec2(i, j);
        vec2  r = vec2(b) + random(p+b)-f;
        float d = dot(r,r);

        if( d < res )
        {
            res = d;
            mr = r;
            mb = b;
        }
    }

    res = 8.0;
    for( int j=-2; j<=2; j++ )
    for( int i=-2; i<=2; i++ )
    {
        ivec2 b = mb + ivec2(i, j);
        vec2  r = vec2(b) + random(p+b) - f;
        float d = dot(0.5*(mr+r), normalize(r-mr));

        res = min( res, d );
    }

    return res;
}

float getBorder( in vec2 p ) {
    float d = voronoiDistance( p );

    return 1.0 - smoothstep(0.0,0.05,d);
}


vec3 voronoiPattern(vec2 p, float gridSize) {

    // Divide plane into square grids of some size, and get cell coordinates as well as coordinates of p relative to the cell
    vec2 scaledP = p / gridSize;
    vec2 cell = floor(scaledP);
    vec2 cellP = fract(scaledP);

    // Each grid cell has a unique hash from it's coordinates

    // Randomize a point in each grid, using its hash as seed

    // Find 9 closest grid cells to p, find the closest points from them

    // Return distance to closest point, and distance to second closest point.  From them, the distance to the edge can also be calculated.
    // In addition, we have the other distances too, perhaps something interesting can be done with distance to 3rd and 4th closet point?

    // Also return the cell hash, it could be used for randomizing content for that cell

    // Might also be interesting to work with variable density (potentially several points in a cell, or subdivided cells?)
    // Could some domain warping be used on the voronoi pattern?
    // Multiple layers could also be used

    // TODO
    return vec3(0.);
}



//===============================================================
//     Material structure and utils
//===============================================================

// Objects can have a material.  The material has the following parameters:
struct Material {    
    vec3 color; // Flat color of material
    float gloss; // Amount of specular light to apply 
    float specularPower; // How small the specular highlight is (indicates smoother surface)
    float reflectivity;
};

// Returns first material when t <= 0, 
// and second when t >= 1, 
// and linearily interpolates between them in the middle.
Material mixMaterials(Material a, Material b, float t) {
    if (t <= 0.) return a;
    else if (t >= 1.0) return b;
    else return Material(
        mix(a.color, b.color, t),
        mix(a.gloss, b.gloss, t),
        mix(a.specularPower, b.specularPower, t),
        mix(a.reflectivity, b.reflectivity, t)
    );
}


//===============================================================
//     Materials in Scene
//===============================================================

const int SNOW = 1;
const int ICE = 2;
const int OCEAN = 3;
const int WATER = 4;
const int SLUSH = 5;
const int ICEBERGS = 6;

// Returns properties of the specified material in the scene
// Material id is the id of the material,
// materialParameter is a floating point parameter that can be used to change the appearance of the material in different palces,
// worldPos is the position in the world, can be used for rendering 3D materials (a checkerboard at the simplest),
// viewDir is the direction the material is viewed from, 
// surfaceNormal is the normal of the material.  Together they can be used to create anisotropic materials, like silk or moss.
Material getMaterial(int materialId, float materialParameter, vec3 worldPos, vec3 viewDir, vec3 surfaceNormal) {

    // Sky is -1, but is handled separately currently.

    /*
    // 0 = Asphalt, param controls lines
    if (materialId == 0) return mixMaterials(
        Material(vec3(0.02, 0.03, 0.05), 0.02, 50.),
        Material(vec3(0.3, 0.3, 0.3), 0.03, 100.), 
        materialParameter
    );
    */


    Material snow = Material(vec3(0.55f, 0.76f, 0.89f), 0.2, 20., 0.0);
    Material water = Material(vec3(0.2f, 0.69f, 0.76f), 0.9, 100., 0.8);
    Material ice = Material(vec3(0.21f, 0.57f, 0.88f), 0.6, 10., 0.07);
    Material iceberg = Material(vec3(0.68f, 0.84f, 0.95f), 0.6, 1., 0.);


    if (materialId == OCEAN) {
        // Mix ocean and icebergs
        return mixMaterials(
            Material(vec3(0.01f, 0.05f, 0.16f), 0.9, 100., 0.), // Sea
            Material(vec3(0.66f, 0.73f, 0.77f), 0.1, 10., 0.), // Ice flake
            materialParameter
        );
    }
    else if (materialId == SNOW) return snow;
    else if (materialId == ICE) return ice;
    else if (materialId == ICEBERGS) return iceberg;
    else if (materialId == WATER) return water;
    else if (materialId == SLUSH) {
        // Mix ocean and icebergs
        return mixMaterials(
            ice, 
            water,
            materialParameter
        );
    }
    else return Material(vec3(1.0f, 0.0f, 0.88f), 0., 1., 0.); // Error material
}


//===============================================================
//     Ocean
//===============================================================

struct OceanParams {
    float height;
};

SceneSample ocean(vec3 p, OceanParams params) {

    // Start with simple plane
    float dist = p.y - params.height;

    float iceBlockScale = 16.0;
    float iceBlockSize = 0.2;
    float iceBlock = 0.;//voronoiDistance(p.xz / iceBlockScale) - iceBlockSize;

    return SceneSample(dist, OCEAN, iceBlock);
}


//===============================================================
//     Dragon
//===============================================================

SceneSample dragonSegment(vec3 p, int segment) {

    // Move space to center on the segment
    vec3 segmentPos = dragonPosition[segment];
    p = move(p, segmentPos);
    p = xRot(p, dragonXrotMod*(sin((time/4.) )+cos((float(segment)/(2.*float(DRAGON_LENGTH))))));
    //p = yRot(p, )

    // Start with simple spheres
    float segmentDiam = 1.0-(float(segment)/(float(DRAGON_LENGTH)*1.5));
    float dist = sdOctahedron(p, segmentDiam)-0.1;

    return SceneSample(dist, ICE, 0.0);    
}

SceneSample dragon(vec3 p) {

    // TODO: Get bounding box from javascript side, and use it as bounds?
    // Move up from ocean when it should appear
    float dragonRestingDepth = 30.0;
    float yOffs = -dragonRestingDepth * (1. -dragonOnScreen);
    p.y -= yOffs;

    // Combine segments
    SceneSample result = dragonSegment(p, 0);
    for (int i = 1; i < DRAGON_LENGTH; i++) {
        result = closest(result, dragonSegment(p, i));
    }

    return result;

}


//===============================================================
//     Snowflake
//===============================================================


struct SnowflakeBranchParams {
    float branchStemLength;
    float branchStemThickness;
    float sideLength;
    float sideThickness;
    float innerLength;
    float innerThickness;
    int sideRepeats;
    

};

struct SnowflakeParams {
    float centerHexSize;
    float snowflakeThickness;
    float stemLength;
    float roothStemThickness; 
    bool sideContinues;
    SnowflakeBranchParams branch1;
    SnowflakeBranchParams outerBranch2;
    SnowflakeBranchParams innerBranch2;
    SnowflakeBranchParams sideBranch2;
};


// Half the stem of a snowflake branch part
float half_stem(vec2 p, float stem_length, float stem_thickness, float symmetry_angle) {
    float half_stem_length = stem_length/2.0;
    float half_stem_thick = stem_thickness/2.0;
    
    float tip_angle = min(radians(180.0 - symmetry_angle), radians(symmetry_angle));
    float outer_edge_length = stem_length - 2.0*stem_thickness / tan(tip_angle);
    
    float flake_stem_half = sdTrapezoid(move2d(p.xy, vec2(half_stem_length, half_stem_thick)), half_stem_length, outer_edge_length/2.0, half_stem_thick);
    return flake_stem_half; 
}

float stem(vec2 p,  float stem_length, float stem_thickness, float symmetry_angle) {

    // Only evaluate the branch on the side we are on
    if (p.y >= 0.0) {
        return half_stem(p, stem_length, stem_thickness, symmetry_angle);
    }
    else {
        return  half_stem(y_mirror2d(p), stem_length, stem_thickness, symmetry_angle);
    }
} 




// A level 2 branch of a snowflake
float branch_2(vec2 p, float symmetry_angle, SnowflakeParams parameters, bool isInner, bool isSide, float prevBranchStemLength) {

    // Parameters

    SnowflakeBranchParams localParams = parameters.outerBranch2;
    if (isInner) {
        localParams = parameters.innerBranch2;
    }
    if (isSide) {
        localParams = parameters.sideBranch2;
    }


    float branchStemLength = localParams.branchStemLength;
    float stemThickness3 = localParams.branchStemThickness;
    float side_length =  localParams.sideLength;
    float side_thickness =  localParams.sideThickness;
    float innerLength = localParams.innerLength;
    float innerThickness = localParams.innerThickness;
    
    // Projection
    vec2 p_tip = move2d(p, vec2(prevBranchStemLength, 0.0));
    vec2 p_tip_half = p_tip;
    if (localParams.sideRepeats >= 2){
        p_tip_half = move2d(p, vec2(prevBranchStemLength/2.0, 0.0));
    }         
    
    
    float result = BIG_ENOUGH;  

    // Stem (so last fromtward branch of flake)
    if (branchStemLength > 0.0){
        result = min(result, stem(p_tip, branchStemLength, stemThickness3, symmetry_angle));    
    }

    //smallest side branches
    // Only evaluate the branch on the side we are on
    if (side_length > 0.0){

  
        if (p.y >= 0.0) {
            // Left
            if (localParams.sideRepeats >= 1){
                vec2 p_left = rotate2d(p_tip, symmetry_angle);
                result = min(result, stem(p_left, side_length, side_thickness, symmetry_angle));
            }
            if (localParams.sideRepeats >= 2){
                vec2 p_left = rotate2d(p_tip_half, symmetry_angle);
                result = min(result, stem(p_left, innerLength, innerThickness, symmetry_angle));
            }
        }
        else {
            // Right
            if (localParams.sideRepeats >= 1){
                vec2 p_right = rotate2d(p_tip, -symmetry_angle);
                result = min(result, stem(p_right, side_length, side_thickness, symmetry_angle));
            }
            if (localParams.sideRepeats >= 2){
                vec2 p_right = rotate2d(p_tip_half, -symmetry_angle);
                result = min(result, stem(p_right, innerLength, innerThickness, symmetry_angle));

            }
        }
    }

    return result;
}

// A branch of a snowflake
float branch(vec2 p, float symmetry_angle, SnowflakeParams parameters) {
    // Parameters
    
    SnowflakeBranchParams localParams = parameters.branch1;
    float branchStemLength = localParams.branchStemLength;
    float stemThickness2 = localParams.branchStemThickness;
    float side_length =  localParams.sideLength;
    float side_thickness =  localParams.sideThickness;
    float innerLength = localParams.innerLength;
    float innerThickness = localParams.innerThickness;
    
    // Projection
    vec2 p_tip = move2d(p, vec2(parameters.stemLength, 0.0));
    vec2 p_tip_half = p_tip;
    if (localParams.sideRepeats >= 2){
        p_tip_half = move2d(p, vec2(parameters.stemLength/2.0, 0.0));
    }    
    
    //draw second stem
    float result = BIG_ENOUGH;

    // Front
    //stem
    result = min(result, stem(p_tip, branchStemLength, stemThickness2, symmetry_angle));
    //branch forward

    if (branchStemLength > 0.0){
        result = min(result, branch_2(p_tip, symmetry_angle, parameters, false, false, branchStemLength));  
  
    }
    

    // Only evaluate the branch on the side we are on
    if (side_length > 0.0){
        if (p.y >= 0.0) {
            // Left

            
            //branches
            if (localParams.sideRepeats >= 1){
                vec2 p_left = rotate2d(p_tip, symmetry_angle);
                result = min(result, stem(p_left, side_length, side_thickness, symmetry_angle));
                if (parameters.sideContinues) {
                    result = min(result, branch_2(p_left, symmetry_angle, parameters, false, true, side_length));
                }
            }
            if (localParams.sideRepeats >= 2){
                vec2 p_left = rotate2d(p_tip_half, symmetry_angle);

                result = min(result, stem(p_left, innerLength, innerThickness, symmetry_angle));         
                if (parameters.innerBranch2.sideRepeats > 0){
                     result = min(result, branch_2(p_left, symmetry_angle, parameters, true, false, innerLength));
                }
                
            
            }    
        
        }
        else {
            // Right

             //branches
            if (localParams.sideRepeats >= 1){
                vec2 p_right = rotate2d(p_tip, -symmetry_angle);
                result = min(result, stem(p_right, side_length, side_thickness, symmetry_angle));

                if (parameters.sideContinues){
                    result = min(result, branch_2(p_right, symmetry_angle, parameters, false, true, side_length));
                }
            }
            if (localParams.sideRepeats >= 2){
                vec2 p_right = rotate2d(p_tip_half, -symmetry_angle);
                result = min(result, stem(p_right,  innerLength, innerThickness, symmetry_angle));
                if (parameters.innerBranch2.sideRepeats > 0){
                     result = min(result, branch_2(p_right, symmetry_angle, parameters, true, false, innerLength));
                }
            }   
        }
    }

    


    return result;
}



// max_radius is calculated by the caller using calculate_snowflake_max_radius
float snowflake_2d(vec2 p, SnowflakeParams parameters, float max_radius) {

    // Check bounding circle
    float margin = 0.1 * max_radius;

    // Check bounding hexagon
    /*float distance_to_bounding_hex = sdHexagon(p, max_radius);
    if (distance_to_bounding_hex > margin) {
        return distance_to_bounding_hex;
    }*/

    // Parameters

    float stemLength = parameters.stemLength;   
    float stem_thickness = parameters.roothStemThickness;
    float center_hex_size = parameters.centerHexSize;
    
    int symmetry = 6;
    int symmetry_angle = 360 / symmetry;
    
    // Hexagon at center of snowflake
    // NOTE: This only looks good with symmetry 6 (and maybe 3 and 12).
    //       Could be refactored later to generic N-gon, or something like that.
    float flake_center_hex = sdHexagon(p, center_hex_size); 
    
    // Check within which sector we are sampling, and only calculate the geometry within that sector.
    // this works, because the sectors are symmetrical, so the closest point must be within the current sector.
    // Speeds up rendering within the snowflake 6 times, because we don't have to take distances to each branch and minimum of those.
    int angle = int(round(degrees(atan(p.y, p.x))));
    int sector = ((angle + 360 + symmetry_angle/2) / symmetry_angle) % symmetry;
    float sectorAngle = float(symmetry_angle * sector);
    vec2 rotated_space_p = rotate2d(p, sectorAngle);
    //only do branches if we have any length

    float wholeBranchDistance = BIG_ENOUGH;
    if (stemLength > 0.0){
        // Stem 
        float stem_distance = stem(rotated_space_p, stemLength, stem_thickness, float(symmetry_angle));    
        //branch
        float branch_distance = branch(
            rotated_space_p, 
            float(symmetry_angle),
            parameters
        );

        wholeBranchDistance =  min(branch_distance, stem_distance);
    }
    // Compose branches and center
    float flake_2d = min(flake_center_hex,wholeBranchDistance);

    return flake_2d;
}



SceneSample snowflake_3d(vec3 p, SnowflakeParams parameters, float materialParameter) {

    // Check bounding sphere
    float max_radius = max( parameters.centerHexSize, parameters.stemLength+parameters.branch1.branchStemLength+parameters.outerBranch2.branchStemLength)+0.1;
    float margin = 0.1 * max_radius;
    /*float distance_to_bounding_sphere = length(p) - max_radius;
    if (distance_to_bounding_sphere > margin) {
        return SceneSample(distance_to_bounding_sphere, SLUSH, 0.);
    }*/

    float thickness = parameters.snowflakeThickness;

    // Check distance from flat snowflake plane
    /*
    float z_surface_distance = abs(p.z) - thickness;
    if (z_surface_distance > margin) {
        return SceneSample(z_surface_distance, SLUSH, 0.); 
    }
    */    


    // Extude flat snowflake into 3d shape

    vec2 p2d = p.xy;    

    float flake_2d = snowflake_2d(p2d, parameters, materialParameter);
    float d = opExtrusion(p, flake_2d, thickness);
    
    return SceneSample(d, SLUSH, materialParameter);
}


SceneSample waterDroplet(vec3 p, float radius, float materialParameter) {
    float distance = length(p) - radius;

    return SceneSample(distance, SLUSH, materialParameter);
}



//===============================================================
//     Scene animation
//===============================================================
/*
SnowflakeParams tuneSnowflakeParameters() {

    // TODO: Randomize parameters from a seed given as parameter, or tune over time


    float size = wobble(4., 2., 1., 741.123);
    float centralHexScale = wobble(6., .1, .1, 312.565);
    float rootStemLengthScale = wobble(8., .6, .2, 672.902);
    float rootStemThicknessScale = wobble(3.3, .1, .06, 289.123);

    SnowflakeBranchParams branch1 = SnowflakeBranchParams(
         wobble(8.5, 0.6, 0.1, 831.512),
         wobble(9.3, 0.6, 0.1, 03.11),
         wobble(6.3, 0.5, 0.1, 92.12),
         wobble(7.2, 0.5, 0.1, 123.12), 
         3,
         true
    );

    SnowflakeBranchParams innerBranch2 = SnowflakeBranchParams(
         wobble(4.5, 0.5, 0.1, 823.13),
         wobble(3.3, 0.5, 0.1, 723.1311),
         wobble(2.3, 0.4, 0.1, 95.234),
         wobble(1.7, 0.4, 0.1, 906.234),
         0,
         false
    );

    SnowflakeBranchParams outerBranch2 = SnowflakeBranchParams(
         wobble(4.5, 0.5, 0.1, 823.13),
         wobble(3.3, 0.5, 0.1, 723.1311),
         wobble(2.3, 0.4, 0.1, 95.234),
         wobble(1.7, 0.4, 0.1, 906.234),
         2,
         true
    );

    SnowflakeParams params = SnowflakeParams(
        size,
        centralHexScale,
        rootStemLengthScale,
        rootStemThicknessScale,
        branch1,
        outerBranch2,
        innerBranch2
    );

    return params;
}
*/


OceanParams calculateOceanParams() {
    float oceanHeight = -4.;

    return OceanParams(
        oceanHeight
    );
}

// Parameters that are used to animate the whole scene / demo.
// Ideally provided from the CPU or vertex shader side
struct SceneParameters {
    //SnowflakeParams mainSnowflake;
    OceanParams ocean;
};

// Do various demo-specific parameter animations, that doesn't depend on which particular pixel we hit
// TODO: This could / should ideally be done on the CPU side, or in the vertex shader.
SceneParameters calculateSceneParameters() {
    return SceneParameters(
        //tuneSnowflakeParameters(),
        calculateOceanParams()
    );
}


SnowflakeParams getSnowflakeParams(int flakeNumber, float hexSizeScale, float stemLength1Scale, float stemLenght2Scale, float stemLenght3Scale,  float stemThicknessScale){
    //all scales from 0-1


    if (flakeNumber == 1 ){

        //hardcoded numbers for flake 1
        float hexSize = 0.5*hexSizeScale;
        float stemLength1 = 1.4*stemLength1Scale;
        float stemLength2 = 1.0*stemLenght2Scale;
        float stemLength3 = 0.7*stemLenght3Scale;
        float baseThickness = 0.1*stemThicknessScale;
        

        return SnowflakeParams(
            hexSize,
            0.05,
            stemLength1,
            baseThickness,
            true,
            SnowflakeBranchParams(
                stemLength2,
                baseThickness*0.8,
                stemLength2*0.7,
                baseThickness*0.7,
                stemLength2*1.1,
                baseThickness*0.75,
                2

            ), 
            SnowflakeBranchParams(
                stemLength3,
                baseThickness*0.5,
                stemLength3*0.5,
                baseThickness*1.2,
                stemLength3,
                baseThickness*0.5,
                2

            ),  
            SnowflakeBranchParams(
                stemLength3,
                0.1,
                stemLength3,
                0.1,
                stemLength3,
                0.1,
                0

            ),
            SnowflakeBranchParams(
                stemLength3*0.8,
                baseThickness*0.6,
                stemLength3*0.4,
                baseThickness*0.6,
                stemLength3,
                baseThickness,
                1

            )
        );
    }
    // add different flakes here with else if
    else if (flakeNumber == 2 ){

        //hardcoded numbers for flake 1
        float hexSize = 0.6*hexSizeScale;
        float stemLength1 = 1.1*stemLength1Scale;
        float stemLength2 = 1.0*stemLenght2Scale;
        float stemLength3 = 1.0*stemLenght3Scale;
        float baseThickness = 0.12*stemThicknessScale;
        

        return SnowflakeParams(
            hexSize,
            0.05,
            stemLength1,
            baseThickness,
            false,
            SnowflakeBranchParams(
                stemLength2,
                baseThickness*0.8,
                stemLength2*0.85,
                baseThickness*0.7,
                stemLength2*0.5,
                baseThickness*2.0,
                2

            ), 
            SnowflakeBranchParams(
                stemLength3,
                baseThickness*0.5,
                stemLength3*0.8,
                baseThickness*0.5,
                stemLength3,
                baseThickness*0.5,
                2

            ),  
            SnowflakeBranchParams(
                stemLength3*0.5,
                baseThickness*2.0,
                stemLength3*0.5,
                0.1,
                stemLength3,
                0.1,
                0

            ),
            SnowflakeBranchParams(
                stemLength3*0.8,
                baseThickness*0.6,
                stemLength3*0.4,
                baseThickness*0.6,
                stemLength3,
                baseThickness*0.7,
                1

            )
        );
    }

    
    else{

        float flakeNumPar1 = mod(float(flakeNumber), 2.0)+1.0;
        float flakeNumPar2 = mod(float(flakeNumber), 6.0)/2.0;
        float flakeNumPar3 = (mod(float(flakeNumber), 4.0)+1.0)/2.0;
        float flakeNumPar4 = flakeNumPar3-1.0;

         //hardcoded numbers for flake 1
        float hexSize = 0.5*hexSizeScale;
        float stemLength1 = (1.0+(flakeNumPar1*0.15))*stemLength1Scale;
        float stemLength2 = (1.0-(flakeNumPar3*0.05))*stemLenght2Scale;
        float stemLength3 = (1.0-(flakeNumPar4*0.2))*stemLenght3Scale;
        float baseThickness = (0.1+(flakeNumPar2*0.02))*stemThicknessScale;


        

        return SnowflakeParams(
            hexSize,
            0.05,
            stemLength1,
            baseThickness,
            true,
            SnowflakeBranchParams(
                stemLength2,
                baseThickness*0.8,
                stemLength2*0.7,
                baseThickness*0.7,
                stemLength2*1.1,
                baseThickness*0.75,
                int(flakeNumPar1)

            ), 
            SnowflakeBranchParams(
                stemLength3,
                baseThickness*0.5,
                stemLength3*(0.5+(flakeNumPar1*0.2)),
                baseThickness-(0.05+(flakeNumPar3*0.02)),
                stemLength3,
                baseThickness*0.5,
                int(flakeNumPar3)

            ),  
            SnowflakeBranchParams(
                stemLength3,
                0.1,
                stemLength3,
                0.1,
                stemLength3,
                0.1,
                int(flakeNumPar4)

            ),
            SnowflakeBranchParams(
                stemLength3*0.8,
                baseThickness*0.6,
                stemLength3*0.4,
                baseThickness*0.6,
                stemLength3,
                baseThickness,
                int(flakeNumPar2)

            )
        );
    }
   


}


SceneSample makeSlush(vec3 p, SnowflakeParams flakeParams, float dropletRadius){

    float max_radius = max( flakeParams.centerHexSize*2., (flakeParams.stemLength+flakeParams.branch1.branchStemLength+flakeParams.outerBranch2.branchStemLength+0.1)*1.5);
    float materialParameter = dropletRadius/max_radius;
    max_radius = max(max_radius, dropletRadius);
    float margin = 0.1 * max_radius;
    float distance_to_bounding_sphere = length(p) - max_radius;
    if (distance_to_bounding_sphere > margin) {
        return SceneSample(distance_to_bounding_sphere, SLUSH, materialParameter);
    }

    SceneSample snowFlake = snowflake_3d(p, flakeParams, materialParameter);

    SceneSample droplet = waterDroplet(p, dropletRadius, materialParameter);
    
    
    return smootClosest(droplet, snowFlake, 0.10);

}

//===============================================================
//     Growng snow scene
//===============================================================

SceneSample growingSnow(vec3 p) {



    // calculate position of the flake
    p = yRot(p, mod(sin(time / 13.1)+ growingBaserotmod , 1.0));
    
    // Move up
    p = up(p, growingDropletYOffset);
    p = xRot(p, growingFlakerot);

    float dropletMaxRadius = 1.0;

    float dropletRadius = growingDropletRadiusScale*dropletMaxRadius;
    
    // return empty if scene is not on
    if (growingSceneOn < 0.9){
        return SceneSample(BIG_ENOUGH, WATER, 0.);
    }

    //otherwise build scene wth values from the controlSignals
    SnowflakeParams flakeParams = getSnowflakeParams(int(growingFlakeNumber), growingHexSizeScale, growingHalfFlakeScale, growingHalfFlakeScale, growingFlakeScale, growingFlakeScale);
    SceneSample flake = makeSlush(p, flakeParams, dropletRadius);

    return flake;
}


SceneSample snowstorm(vec3 p) {

    if (snowstormStart < 0.1){// no snowstorm
        return SceneSample(BIG_ENOUGH, WATER, 0.);
    }
    float snowstormtime = time-snowstormStart;

    SceneSample storm = SceneSample(BIG_ENOUGH, WATER, 0.);
    for (int i = 0; i < 5; i++) {
        float sif = float(i);
        float offsetparam = (-(snowstormtime+(sif*4.0))/(0.6+0.2*sif));
        float offsetmod = mod(offsetparam, 25.0);
        float flakesize= stormStartScale*min((25.0-offsetmod)/3.0, 1.0);
        float flakenumber = sif + floor(offsetparam/25.0);
        float yOffset = offsetmod-6.0;
        vec3 p_t = up(p, yOffset);
        p_t = back(p_t, mod(sif*23.0, 40.0)- 20.0);
        p_t = right(p_t,(mod((5.0-sif)*37.0, 40.0)- 20.0)+5.0*sin(snowstormtime/2.0));
        p_t = xRot(p_t, 0.2+cos((time+flakenumber)/5.0));
        SnowflakeParams flakeParams = getSnowflakeParams(int(flakenumber), flakesize, flakesize, flakesize, flakesize, flakesize);
        SceneSample flake = makeSlush(p_t, flakeParams, 0.);
        storm = closest(storm, flake);

    }
    return storm;
    
    //return SceneSample(BIG_ENOUGH, WATER, 0.);
}


//===============================================================
//     Ice Cube
//===============================================================

SceneSample iceCube(vec3 p) {

    // Ice block global scale
    float iceCubeScale = 4.0;

    // Ice cell size
    float areaSizeX = iceCubeScale * 7.3;
    float areaSizeZ = iceCubeScale * 8.3;

    // Divide ground into squares
    int x = int(round(p.x / areaSizeX));
    int z = int(round(p.z / areaSizeZ));
    ivec2 cellId = ivec2(x, z);

    // Ice cube size
    float sizeScale = 1.0*iceCubeScale;
    vec3 cubeSize = sizeScale * vec3(
        1. + .2 * random(cellId, 1.), 
        1.3 + .2 * random(cellId, 2.), 
        .7 + .1 * random(cellId, 3.)
    );

    vec3 mov = abs(hash3(uvec3(round(p.x / areaSizeX), round(p.z / areaSizeZ), 0)));

    p.x = p.x - round(p.x / areaSizeX) * areaSizeX;
    p.z = p.z - round(p.z / areaSizeZ) * areaSizeZ;
    p.x += mov.x * 2.;
    p.z += mov.y * 2.;
    p.y += 4.;

    // Move ice cube
    float scatterAmount = 0.3;
    p.x += scatterAmount * areaSizeX * (random(cellId, 7.) - 0.5);
    p.z += scatterAmount * areaSizeZ * (random(cellId, 8.) - 0.5);

    // Rotate ice cube
    p = xRot(p,  random(cellId, 4.));
    p = yRot(p,  random(cellId, 5.));
    p = zRot(p,  random(cellId, 6.));

    // Box equation
    vec3 q = abs(p) - cubeSize;
    float dist = length(max(q, 0.0)) + min(max(q.x, max(q.y, q.z)), 0.0) - 0.;


    return SceneSample(dist, ICEBERGS, 0.0);
}



//===============================================================
//     Scene
//===============================================================


// Sample the scene at the specified point in space.
// Returns the (approximate, lower bound) distance to the closest surface in x,
// the material id in y, and a material parameter in z.
SceneSample scene(vec3 p, SceneParameters parameters) {

    // Simple rotation
    //p = yRot(p, sin(time / 100.1) * 5.0 );

    // Move up
    //p = up(p, 0.);

    // Growing snow
    SceneSample result = growingSnow(p);

    result = closest(result, snowstorm(p));

    // Ocean
    result = closest(result, ocean(p, parameters.ocean));

    // Ice cube
    result = closest(result, iceCube(p));

    // Ball
    //result = closest(result, SceneSample(length(up(p, 7.)) - 2., ICE, 0.0));

    // Dragon
    if (dragonOnScreen > 0.01) {
        result = closest(result, dragon(p));
    }

    // Return the closest one
    return result;
}  



//===============================================================
//     Rendering
//===============================================================

// Calculate a normal at the specified point in the scene, that points away from the nearest surface.
vec3 sceneNormal(vec3 pos, float cameraDist, SceneParameters sceneParameters) {
    // Center sample
    float c = scene(pos, sceneParameters).surfaceDistance;

    // Use offset samples to compute gradient / normal
    float epsilon = mapClamp(cameraDist, 1., 1000., 0.01, 1.);
    vec2 eps_zero = vec2(epsilon, 0.0);
    return normalize(vec3( 
        scene(pos + eps_zero.xyy, sceneParameters).surfaceDistance, 
        scene(pos + eps_zero.yxy, sceneParameters).surfaceDistance, 
        scene(pos + eps_zero.yyx, sceneParameters).surfaceDistance 
    ) - c);
}


// Given a direction normal, calculates the color of the sky in that direction
vec3 skyColorInDir(vec3 dir) {
    float atmo = pow(max(0., 1. - dir.y), 3.);
    float hor = pow(max(0., 1. - dir.y), 10.);
    float sunDot = dot(sunDir, dir);
    float sun = pow(max(0.,sunDot), 3.);
    float sunAura1 = pow(max(0.,sunDot), 5.);
    float sunAura2 = pow(max(0.,sunDot), 30.);
    float sunDisc = pow(max(0.,sunDot), 1000.);
    float backLight = pow(max(0.,-sunDot), 1.);
    float shadow = pow(max(0., -dir.y), 1.2);

    vec3 col = vec3(0);
    col += vec3(0.0157, 0.0196, 0.0941);
    col += atmo * vec3(0.0235, 0.0863, 0.2745);
    col += hor * vec3(0.1765, 0.0667, 0.2314);
    col += sunAura1 * vec3(0.3686, 0.149, 0.1176);
    col += sunAura2 * vec3(0.5137, 0.251, 0.1176);
    col += 2.*sunDisc * vec3(0.9843, 0.4706, 0.3137);
    col += atmo * sun * vec3(0.3608, 0.1255, 0.0784);
    col += atmo * backLight * vec3(0.1804, 0.0784, 0.1961);
    col -= col*shadow * 0.5 * vec3(0.9137, 0.7294, 0.6235);
    col *= (1. - shadow*0.5);

    return col;
}



// Calculates fog color with a specified distance to the target, and direction of ray.
// Returns color in rgb, and amount of fog in a.
vec4 calculateFog(float dist, // camera to point distance
                  vec3 cameraPos,
                  vec3 rayDir) {


    // Some handwavy height fog
    float horizon = 1. - max(dot( rayDir, vec3(0.,1.,0.) ), 0.);
    float atmospherePassed = dist * (0.02 + pow(horizon, 6.));

    float fogAmount = 1.0 - exp( -atmospherePassed * fogIntensity );

    float shadow = pow(min(1.,1.+rayDir.y),1.);

    vec3 fogColor = fogAmount * shadow * skyColorInDir(rayDir); 
    return vec4(fogColor, fogAmount);
}




// Based on IQ:s method improved by Sebastian Aaltonen: https://iquilezles.org/articles/rmshadows/
float sceneShadow(in vec3 rayOrigin, in vec3 lightDir, float sharpness, vec3 surfaceNormal, SceneParameters sceneParameters) {

    float prevH = 1e20;

    float hitThreshold = 0.01;

    // Shade when this close to backface.  If we had proper hightlights this would probably not be needed.
    const float shadowsEdge = 0.1;

    // If the surface is already shadowed, return
    float surfaceDotLight = dot(lightDir, surfaceNormal);
    if (surfaceDotLight <= 0.) return 0.;
    //else return 1.;

    //return max(0., surfaceDotLight);

    // Lift start of shadow ray a bit from surface, reduces banding and smoothens sharpness of shadow.
    rayOrigin += surfaceNormal * 0.1;


    // Shadow is 0 when full shadow, 1 when no shadow.

    float shadow = mapClamp(surfaceDotLight, 0., shadowsEdge, 0., 1.) ; // Shade when approaching backface
    float t = 0.1; // startDistance
    for (int i = 0; i < MAX_SHADOW_RAY_STEPS; i++) {

        vec3 p = rayOrigin + lightDir * t;
        float stepDist = scene(p, sceneParameters).surfaceDistance * 0.95;

        // If we hit an object, we are shadowed
        if( stepDist <= hitThreshold ) return 0.0;

        // Shadow based on how close to any object we traverse
        float y = stepDist * stepDist / (2.0 * prevH);
        float d = sqrt(stepDist * stepDist - y * y);
        shadow = min(shadow, sharpness * d / max(0.0001, t - y));
        shadow = max(0., shadow);

        prevH = stepDist;

        //shadow = min( shadow, sharpness * h/t );
        t += stepDist;

        if (t >= MAX_DISTANCE) return shadow;
    }

    return shadow;
}


// Used to return the rendering results of a pixel
struct RenderingResult {
    vec3 lightColor;
    float complexity; // Amount of raymarching steps done, 0=none, 1=max allowed amount.  Useful for debugging and special effects.
};

struct RayCastResult {
    SceneSample sceneData;
    float complexity;
    vec3 p;
    bool skyHit;
    vec3 rayDir;
};


RayCastResult castRay(in vec3 rayOrigin, in vec3 rayDir, inout float rayDist, out int iterations, inout vec3 col, inout float nextColPrc, SceneParameters sceneParameters) {

    // ------------------------ Cast Ray ------------------------ 

    SceneSample sceneData; // The distance to the objects in the scene.  Also include object material id and a parameter.
    SceneSample prevSceneData = scene(rayOrigin, sceneParameters); 
    for (int i = 0; i < MAX_RAY_STEPS; i++) {

        // Calculate position
        vec3 p = rayOrigin + rayDir * rayDist;

        // Sample scene at current position
        sceneData = scene(p, sceneParameters);
        float surfaceDistance = sceneData.surfaceDistance;
        float previousSurfaceDistance = prevSceneData.surfaceDistance;

        // Move along ray
        float stepDist = surfaceDistance * STEP_DISTANCE_MULTIPLIER; // Take a bit shorter steps as we have non-strict distance functions.
        rayDist += stepDist;

        // Count iterations (in theory i could be used, but it isn't in scope after the loop?)
        iterations++;

        // Check if we hit a surface
        // Use a constant multiplied with distance traveled so far -> less accuracy far away, more close by.
        if (surfaceDistance < (SURFACE_DISTANCE * rayDist)) { 

            // Interpolate hitpoint based on previous and current distance
            if (surfaceDistance >= 0.) {
                // Stopped outside surface, step forward until we approach surface edge.
                rayDist += surfaceDistance;
            }
            else if (previousSurfaceDistance < 0.) {
                // Return to start if we already were inside landscape (camera in landscape?)
                rayDist -= stepDist;
            }
            else {
                // Surface distance < 0 and previous surface distance >= 0
                // We ran into surface.
                // Take average of previous and current surface distance to get an approximation of where the surface is, and step back there
                float stepBackLength = 0.5 * (surfaceDistance + (stepDist - previousSurfaceDistance));
                rayDist -= stepBackLength;
                
                // This seems wrong now?:
                ////, step back proportionally to how far the current distance is compared to the previous plus current.
                ////rayDist -= stepDist * (-currentDist) / (prevDist - currentDist);

            }

            // Normal of point we hit
            vec3 n = sceneNormal(p, rayDist, sceneParameters);

            // ------------ Material ------------

            // Get material properties of the the point we hit
            Material material = getMaterial(sceneData.materialId, sceneData.materialParameter, p, rayDir, n);

            if (material.reflectivity > 0.0) {
                rayDir = rayDir - 2. * dot(rayDir, n) * n;
                rayOrigin = p + rayDir * .5;
                col += nextColPrc * (1. - material.reflectivity) * material.color;
                nextColPrc *= material.reflectivity;
            } else {
                // Ray stopped
                break;
            }
        }

        // Check if we went off into the sky
        if (rayDist >= MAX_DISTANCE) {
            sceneData.surfaceDistance = MAX_DISTANCE;
            sceneData.materialId = -1;
            rayDist = MAX_DISTANCE;
            break;
        }

        prevSceneData = sceneData;
    }
    
    // Indicates how many iterations were used, 1 = max, 0 = min.
    float complexity = float(iterations) / (float(MAX_RAY_STEPS) - 1.0);

    // Position of point we hit
    vec3 p = rayOrigin + rayDir * rayDist;

    // Determine if we hit sky (either directly, or if we got very close)
    bool skyHit = rayDist >= MAX_DISTANCE || (iterations >= MAX_RAY_STEPS - 1 && rayDist > MAX_DISTANCE * 0.99);

    return RayCastResult(sceneData, complexity, p, skyHit, rayDir);
}


// Raymarches from the camera in the direction of the ray, and calculates incoming light from the direction.
// Also returns complexity, that reflects the amount of steps done.
RenderingResult render(in vec3 rayOrigin, in vec3 rayDir, SceneParameters sceneParameters) {
    // Determine color
    vec3 col = vec3(0., 0., 0.);
    float nextColPrc = 1.;

    // Cast ray
    float rayDist = 0.1; // Start a bit outside the camera.  Allows camera to drive through objects
    int iterations = 0;
    
    // ------------------------ Render ------------------------ 

    RayCastResult ray = castRay(rayOrigin, rayDir, rayDist, iterations, col, nextColPrc, sceneParameters);

    // Calculate fog effect
    vec4 fogColorAndAlpha = calculateFog(rayDist, rayOrigin,  ray.rayDir);

    if (ray.skyHit) {
        // Hit the sky, get the color of the sky
        col += nextColPrc * skyColorInDir(ray.rayDir);
    }
    else if (fogColorAndAlpha.a >= fogExtinctionThreshold) {
        // If fog effectovely covers surface details, there is no need to calculate the surface things.
        col += nextColPrc * fogColorAndAlpha.rgb;
    }
    else {
        // Normal of point we hit
        vec3 n = sceneNormal(ray.p, rayDist, sceneParameters);

        // ------------ Material ------------

        // Get material properties of the the point we hit
        Material material = getMaterial(ray.sceneData.materialId, ray.sceneData.materialParameter, ray.p, ray.rayDir, n);

        // Specular reflections (not affected by material color)
        vec3 specularLight = vec3(0.);

        // Material of surface
        vec3 surfaceColor = material.color;
        float gloss = material.gloss; // Amount of specular light to apply
        float specularPower = material.specularPower; // Exponent of specular highlight


        // ------------ Sunlight & Sun shadow ------------

        /* We won't need shadows for now, they also kill the performance
        // Sun shadow
        float sunShadow = sceneShadow(p, sunDir, 5., n);

        // Colorize the penumbra for a bit of shadow oomph
        vec3 colorizedSunShadow = pow( vec3(sunShadow), sunShadowPower );
        */

        // Direct light
        vec3 sunLight = sunIntensity * sunColor * max(dot(n, sunDir), 0.0);
        /*
        sunLight *= colorizedSunShadow; // Sun shadow
        */

        // Specular sun reflection
        float sunSpecular = max(0., pow(dot(n, normalize(sunDir - ray.rayDir)), specularPower));
        specularLight += gloss * sunSpecular * sunIntensity * sunColor;
        /*
        specularLight += gloss * sunSpecular * sunIntensity * sunColor * colorizedSunShadow;
        */

        // ------------ Ambient / top-down skylight and shadow ------------

        // Sky light, from straight above
        vec3 skyLight = skyIntensity * skyColor * clamp( 0.5 + 0.5 * n.y, 0.0, 1.0 );
        /*
        float skyShadow = sceneShadow(p, vec3(0., 1., 0.),  2., n);                
        */
        /*
        float fakeAmbientOcclusion = clamp(pow(1. - complexity, 4.), 0., 1.); // Tune with iteration step data
        */
        /*
        skyLight *= skyShadow * 0.25 + 0.75;
        */
        /*
        skyLight *= fakeAmbientOcclusion;
        */

        // Indirect light
        vec3 indirectDir = normalize(sunDir * vec3(-1.0, 0., -1.0)); // Reverse sun direction, and snap to horizontal
        vec3 indirectLight = indirectIntensity * indirectColor * clamp( dot( n, indirectDir), 0.0, 1.0 );
        /*
        indirectLight *= fakeAmbientOcclusion;
        */


        // ------------ Sunlight & Sun shadow ------------

        // Total incloming light to this surface
        vec3 incomingLight = sunLight + indirectLight + skyLight;

        // Diffuse color
        vec3 diffuse = surfaceColor * incomingLight;

        col += nextColPrc * (diffuse + specularLight);

    }


    // ------------ Apply fog ------------

    // This is applied both in case we hit the sky, fogged out by distance, or hit an object

    // Attenuate background color with fog intensity (amount of background light blocked by fog),
    // and add accumulated fog color (light diffused in fog and directed our way).
    // One reduces light and the other adds diffuse light.
    col = col * (1. - fogColorAndAlpha.a) + fogColorAndAlpha.rgb;


    // Return the color from the rendering function, as well as complexity of the render.
    return RenderingResult(col, ray.complexity);
}


//===============================================================
//     Main and Final post processing
//===============================================================


void main() {
    vec3 camTarget = vec3(sin(time * 0.27+.2), sin(time * 0.8)*0.3 + 6., 0); // TODO: Position of target at time
    vec3 camPos = vec3(sin(time*0.12) * 12., 4.4, cos(time*0.11) * 5. -4.); // TODO: cameraPos(time, camTarget);

    // Do scene animation that doesn't depend on specific part of it that is visible in the pixel we are rendering
    SceneParameters sceneParameters = calculateSceneParameters();

    // project screen coordinate into world
	vec2 uv = screenToWorld(gl_FragCoord.xy);

    // Direction that the current fragment is in
    vec3 rayDir = getCameraRayDir(uv, camPos, camTarget);

    // Calculate fragment color from scene.
    RenderingResult result = render(camPos, rayDir, sceneParameters);
    vec3 col = result.lightColor;

    /*
    // Halo effect for debugging
    float stepCountEffect = pow(result.complexity, 1.5);
    col += 2.0 * vec3(0.0, 0.4, 0.5) * stepCountEffect;
    */

    // High dynamic range
    col = col / (col + vec3(1.0));

    // Contrast
    const float contrast = 0.3;
    col = mix(col, smoothstep(0.0, 1.0, col), contrast);

    // Apply fade
    col *= fadeFromBlack;

    // Gamma correct the color
    col = pow(col, vec3(1./2.2));
    
    // Assign to fragment
    FragColor = vec4(col, 1.0);
    
    //gl_FragColor = vec4(sin(time) * test() * p.y, 0.4 * sin(time-1.0) * p.x*p.y, sin(time+1.0) * p.x, 1);
    //FragColor = vec4(1.0, 0.4, 0.0, 1.0);
}


